{ *********************************************************************** }
{                                                                         }
{    Pe ļ Reloc                                                }
{                                                                         }
{   ƣLsuper 2005.12.27                                               }
{   ע                                                                }
{   ˣ                                                                }
{   ˵refact form jrsoftware.org stripreloc 1.13                      }
{                                                                         }
{   Copyright (c) 1998-2005 Super Studio                                  }
{                                                                         }
{ *********************************************************************** }
{
  StripReloc v1.13
  Strip relocation section from Win32 PE files
  Copyright (C) 1999-2005 Jordan Russell. All rights reserved.
  http://www.jrsoftware.org/

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $jrsoftware: stripreloc/StripReloc.dpr,v 1.16 2005/02/13 05:03:42 jr Exp $
}

unit StripReloc;

interface

uses
  SysUtils, Windows, Classes;

type
  TPESectionHeaderArray = array[0..$7FFFFFFF div
    SizeOf(TImageSectionHeader) - 1] of TImageSectionHeader;
  PPESectionHeaderArray = ^TPESectionHeaderArray;

  TStripReloc = class(TObject)
  private
    FFile: TFileStream;
    FForce: Boolean;
    FSetChecksum: Boolean;

    FPEHeaderOffset,
    FPESig: Cardinal;
    FPEHeader: TImageFileHeader;
    FPEOptHeader: TImageOptionalHeader;
    FPESectionHeaders: PPESectionHeaderArray;
    FRelocVirtualAddr,
    FRelocPhysOffset,
    FRelocPhysSize: Cardinal;
  private
    function  CalcMemoryCheckSum(AMemory: Pointer;
      ACount: DWord): DWord;
    function  CalcFileChecksum(AHandle: THandle): DWord;

    function  ParsePeFile: Boolean;
  public
    constructor Create(const AFile: string);
    destructor Destroy; override;

    class function Execute(const AFile, ASave: string;
      ASetChecksum: Boolean = True): Boolean;

    function  SaveToFile(const ASave: string): Boolean;

    property  SetChecksum: Boolean
      read FSetChecksum write FSetChecksum;
    property  Force: Boolean
      read FForce write FForce;
  end;

implementation

{ TStripReloc }

////////////////////////////////////////////////////////////////////////////////
//: Lsuper 2005.12.27
//:  PE ļУ
//
////////////////////////////////////////////////////////////////////////////////
function TStripReloc.CalcFileChecksum(AHandle: THandle): DWord;
var
  pData: Pointer;
  nSize: Cardinal;
  hMemFile: THandle;
begin
  Result := 0;
  nSize := GetFileSize(AHandle, nil);
  hMemFile := CreateFileMapping(AHandle, nil, PAGE_READONLY, 0, nSize, nil);
  if hMemFile <> 0 then
  try
    pData := MapViewOfFile(hMemFile, FILE_MAP_READ, 0, 0, nSize);
    Result := CalcMemoryCheckSum(pData, nSize);
  finally
    if pData <> nil then
      UnmapViewOfFile(pData);
    CloseHandle(hMemFile);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//: Lsuper 2005.12.27
//: ڴУ
//
////////////////////////////////////////////////////////////////////////////////
function TStripReloc.CalcMemoryCheckSum(AMemory: Pointer; ACount: DWord): DWord;
type
  TCheckSumMappedFile = function(BaseAddress: Pointer; FileLength: DWord;
    var HeaderSum, CheckSum: DWord): PImageNtHeaders; stdcall;
var
  nHandle: THandle;
  nHeader: DWord;
  pCalcProc: TCheckSumMappedFile;
begin
  Result := 0;
  if AMemory = nil then
    Exit;
  nHandle := LoadLibrary('imagehlp.dll');
  if nHandle <> 0 then
  try
    @pCalcProc := GetProcAddress(nHandle, 'CheckSumMappedFile');
    if not Assigned(pCalcProc) then
      Exit;
    if pCalcProc(AMemory, ACount, nHeader, Result) = nil then
      Result := 0;
  finally
    FreeLibrary(nHandle);
  end;
end;

constructor TStripReloc.Create(const AFile: string);
begin
  FFile := TFileStream.Create(AFile, fmOpenRead or fmShareDenyWrite);
  FForce := False;
  FSetChecksum := True;
end;

destructor TStripReloc.Destroy;
begin
  if FPESectionHeaders <> nil then
    FreeMem(FPESectionHeaders);
  FreeAndNil(FFile);
  
  inherited;
end;

////////////////////////////////////////////////////////////////////////////////
//: Lsuper 2005.12.27
//: ѹ PE ļ
//
////////////////////////////////////////////////////////////////////////////////
class function TStripReloc.Execute(const AFile, ASave: string;
  ASetChecksum: Boolean): Boolean;
begin
  with Self.Create(AFile) do
  try
    SetChecksum := ASetChecksum;
    Result := ParsePeFile and SaveToFile(ASave);
  finally
    Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//: Lsuper 2005.12.27
//: ļ
//
////////////////////////////////////////////////////////////////////////////////
function TStripReloc.ParsePeFile: Boolean;
const
  IMAGE_NT_OPTIONAL_HDR32_MAGIC = $10B;
var
  nSize,
  I: Integer;
  nExeSig: Word;
begin
  Result := False;

  FRelocPhysOffset := 0;
  FRelocPhysSize := 0;
  if FPESectionHeaders <> nil then
  begin
    FreeMem(FPESectionHeaders);
    FPESectionHeaders := nil;
  end;

  with FFile do
  begin
    Read(nExeSig, SizeOf(nExeSig));
    if nExeSig <> $5A4D {'MZ'} then { File isn't a PE file }
      Exit;
    Position := $3C;
    Read(FPEHeaderOffset, SizeOf(FPEHeaderOffset));
    if FPEHeaderOffset = 0 then { File isn't a PE file }
      Exit;
    Position := FPEHeaderOffset;
    Read(FPESig, SizeOf(FPESig));
    if FPESig <> $00004550 {'PE'#0#0} then { File isn't a PE file }
      Exit;
    Read(FPEHeader, SizeOf(FPEHeader));
  end;
  with FPEHeader do
  begin
    if not FForce and (Characteristics and IMAGE_FILE_DLL <> 0) then { Can't strip a DLL }
      Exit;
    if Characteristics and IMAGE_FILE_RELOCS_STRIPPED <> 0 then { Relocations already stripped from file }
      Exit;
    Characteristics := Characteristics or IMAGE_FILE_RELOCS_STRIPPED;
    if SizeOfOptionalHeader <> SizeOf(FPEOptHeader) then { File isn't a valid 32-bit image }
      Exit;
  end;
  FFile.Read(FPEOptHeader, SizeOf(FPEOptHeader));
  if FPEOptHeader.Magic <> IMAGE_NT_OPTIONAL_HDR32_MAGIC then { File isn't a valid 32-bit image }
    Exit;
  with FPEOptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC] do
  begin
    if (VirtualAddress = 0) or (Size = 0) then { Relocations already stripped from file }
      Exit;
    FRelocVirtualAddr := VirtualAddress;
    VirtualAddress := 0;
    Size := 0;
  end;
  nSize := FPEHeader.NumberOfSections * SizeOf(TImageSectionHeader);
  GetMem(FPESectionHeaders, nSize);
  FFile.Read(FPESectionHeaders^, nSize);
  for I := 0 to FPEHeader.NumberOfSections - 1 do
    with FPESectionHeaders[I] do
      if (VirtualAddress = FRelocVirtualAddr) and (SizeOfRawData <> 0) then
      begin
        FRelocPhysOffset := PointerToRawData;
        FRelocPhysSize := SizeOfRawData;
        SizeOfRawData := 0;
        Break;
      end;
  if FRelocPhysOffset = 0 then { Relocations already stripped from file }
    Exit;
  if FRelocPhysSize = 0 then { Relocations already stripped from file }
    Exit;
  for I := 0 to FPEHeader.NumberOfSections - 1 do
    with FPESectionHeaders[I] do
    begin
      if PointerToRawData > FRelocPhysOffset then
        Dec(PointerToRawData, FRelocPhysSize);
      if PointerToLinenumbers > FRelocPhysOffset then
        Dec(PointerToLinenumbers, FRelocPhysSize);
      if PointerToRelocations <> 0 then { Cannot handle this file }
        Exit;
    end;
  if FPEOptHeader.ImageBase < $400000 then { Cannot handle this file -- the image base address is less than 0x400000 }
    Exit;
  Result := True;
end;

////////////////////////////////////////////////////////////////////////////////
//: Lsuper 2005.12.27
//: 浽ļ
//
////////////////////////////////////////////////////////////////////////////////
function TStripReloc.SaveToFile(const ASave: string): Boolean;
var
  stream: TFileStream;
  tStamp: TFileTime;
begin
  stream := TFileStream.Create(ASave, fmCreate or fmOpenWrite or fmShareExclusive);
  with stream do
  try
    FFile.Position := 0;
    CopyFrom(FFile, FRelocPhysOffset);

    FFile.Seek(FRelocPhysSize, soFromCurrent);
    CopyFrom(FFile, FFile.Size - FFile.Position);

    Position := FPEHeaderOffset + SizeOf(FPESig);
    Write(FPEHeader, SizeOf(FPEHeader));

    if not FSetChecksum then
      FPEOptHeader.CheckSum := 0;
    Write(FPEOptHeader, SizeOf(FPEOptHeader));
    Write(FPESectionHeaders^, FPEHeader.NumberOfSections * SizeOf(TImageSectionHeader));

    if FSetChecksum then { go back and rewrite opt. header with new checksum }
    begin
      FPEOptHeader.CheckSum := CalcFileChecksum(stream.Handle);
      Position := FPEHeaderOffset + SizeOf(FPESig) + SizeOf(FPEHeader);
      Write(FPEOptHeader, SizeOf(FPEOptHeader));
    end;

    GetFileTime(FFile.Handle, nil, nil, @tStamp);
    SetFileTime(Handle, nil, nil, @tStamp);

    Result := True;
  finally
    Free;
  end;
end;

end.
